from typing import Dict, List
from rekuest.api.schema import WidgetInput
from rekuest.definition.registry import get_default_definition_registry
from rekuest.structures.registry import StructureRegistry
from rekuest.structures.default import get_default_structure_registry
from rekuest.actors.types import Actifier
import os
from functools import wraps


def register(
    *func,
    actifier: Actifier = None,
    interface: str = None,
    widgets: Dict[str, WidgetInput] = None,
    interfaces: List[str] = [],
    on_provide=None,
    on_unprovide=None,
    structure_registry: StructureRegistry = None,
    **actifier_params,
):
    """Register a function or actor to the default definition registry.

    You can use this decorator to register a function or actor to the default
    definition registry. There is also a function version of this decorator,
    which is more convenient to use.

    Example:
        >>> @register
        >>> def hello_world(string: str):

        >>> @register(interface="hello_world")
        >>> def hello_world(string: str):

    Args:
        function_or_actor (Union[Callable, Actor]): The function or Actor
        builder (ActorBuilder, optional): An actor builder (see ActorBuilder). Defaults to None.
        package (str, optional): The package you want to register this function in. Defaults to standard app package    .
        interface (str, optional): The name of the function. Defaults to the functions name.
        widgets (Dict[str, WidgetInput], optional): A dictionary of parameter key and a widget. Defaults to the default widgets as registered in the structure registry .
        interfaces (List[str], optional): Interfaces that this node adheres to. Defaults to [].
        on_provide (Callable[[ProvisionFragment], Awaitable[dict]], optional): Function that shall be called on provide (in the async eventloop). Defaults to None.
        on_unprovide (Callable[[], Awaitable[dict]], optional): Function that shall be called on unprovide (in the async eventloop). Defaults to None.
        structure_registry (StructureRegistry, optional): The structure registry to use for this Actor (used to shrink and expand inputs). Defaults to None.
    """
    if len(func) > 1:
        raise ValueError("You can only register one function or actor at a time.")
    if len(func) == 1:
        function_or_actor = func[0]

        @wraps(function_or_actor)
        def wrapped_function(*args, **kwargs):
            return function_or_actor(*args, **kwargs)

        get_default_definition_registry().register(
            function_or_actor,
            structure_registry=structure_registry or get_default_structure_registry(),
            actifier=actifier,
            interface=interface,
            widgets=widgets,
            interfaces=interfaces,
            on_provide=on_provide,
            on_unprovide=on_unprovide,
            **actifier_params,
        )

        return wrapped_function

    else:

        def real_decorator(function_or_actor):
            # Simple bypass for now
            def wrapped_function(*args, **kwargs):
                return function_or_actor(*args, **kwargs)

            get_default_definition_registry().register(
                function_or_actor,
                structure_registry=structure_registry
                or get_default_structure_registry(),
                actifier=actifier,
                interface=interface,
                widgets=widgets,
                interfaces=interfaces,
                on_provide=on_provide,
                on_unprovide=on_unprovide,
                **actifier_params,
            )

            return wrapped_function

        return real_decorator


def register_structure(cls, **kwargs):
    """Register a structure to the default structure registry.

    Args:
        cls (Structure): The structure class
        name (str, optional): The name of the structure. Defaults to the class name.
    """
    get_default_structure_registry().register_as_structure(cls, **kwargs)
    return cls


def create_arkitekt_folder(with_cache: bool = True):
    """Create the arkitekt folder"""
    os.makedirs(".arkitekt", exist_ok=True)
    if with_cache:
        os.makedirs(".arkitekt/cache", exist_ok=True)

    gitignore = os.path.join(".arkitekt", ".gitignore")
    dockerignore = os.path.join(".arkitekt", ".dockerignore")
    if not os.path.exists(gitignore):
        with open(gitignore, "w") as f:
            f.write(
                "# Hiding Arkitekt Credential files from git\n*.json\n*.temp\ncache/"
            )
    if not os.path.exists(dockerignore):
        with open(dockerignore, "w") as f:
            f.write(
                "# Hiding Arkitekt Credential files from git\n*.json\n*.temp\ncache/"
            )

    return ".arkitekt"



from importlib import import_module
from arkitekt.cli.types import Manifest, Packager
from arkitekt.utils import create_arkitekt_folder
from pydantic import BaseModel, Field
import sys
import rich_click as click
from enum import Enum

try:
    from rekuest.api.schema import DefinitionInput
    from rekuest.definition.registry import get_default_definition_registry
except ImportError as e:
    raise ImportError("Please install rekuest to use this feature") from e


import os
from typing import Dict, List, Optional
import yaml
import json
import datetime

from subprocess import check_call


def import_deployer(builder):
    module_path, function_name = builder.rsplit(".", 1)
    module = import_module(module_path)
    function = getattr(module, function_name)
    return function


def generate_definitions(module_path) -> List[DefinitionInput]:
    import_module(module_path)
    reg = get_default_definition_registry()
    return list(reg.definitions.keys())


class Build(BaseModel):
    manifest: Manifest
    build_id: str
    builder: str
    build_at: datetime.datetime = Field(default_factory=datetime.datetime.now)


class ConfigFile(BaseModel):
    builds: List[Build] = Field(default_factory=list)
    latest_build: Optional[Build]


PIP_NO_GPU = """
FROM jhnnsrs/vanilla_pip:latest

# setup conda virtual environment
COPY ./requirements.txt /tmp/requirements.txt
RUN pip install /tmp/requirements.yml

"""

PIP_GPU = """
FROM jhnnsrs/vanilla_pip:latest

# setup conda virtual environment
COPY ./requirements.txt /tmp/requirements.txt
RUN pip install /tmp/requirements.yml

"""


CONDA_GPU = """
FROM jhnnsrs/tensorflow_conda_gpu:latest

# setup conda virtual environment
COPY ./requirements.yml /tmp/requirements.yml
RUN conda env create --name camera-seg -f /tmp/requirements.yml

RUN conda activate camera-seg   
"""


CONDA_NO_GPU = """
FROM jhnnsrs/tensorflow_conda:latest

# setup conda virtual environment
COPY ./requirements.yml /tmp/requirements.yml
RUN conda env create --name camera-seg -f /tmp/requirements.yml

RUN conda activate camera-seg
"""


def build_dockerfile(packager: Packager, gpu: bool, python_version: str) -> str:
    if packager == Packager.CONDA:
        if gpu:
            return CONDA_GPU
        else:
            return CONDA_NO_GPU

    if packager == Packager.PIP:
        if gpu:
            return PIP_GPU
        else:
            return PIP_NO_GPU

    raise click.ClickException(
        f"Packager {packager} {gpu and 'with GPU Support'} is not supported. Please create a issue on github and create your own Dockerfile for now."
    )


class DockerFile(BaseModel):
    base: Optional[str] = None
    commands: List[str] = Field(default_factory=list)


def docker_file_wizard(manifest: Manifest, auto: bool = True):
    # get python version
    python_version = sys.version_info
    python_version = (
        f"{python_version.major}.{python_version.minor}.{python_version.micro}"
    )
    python_version = (
        python_version
        if auto
        else click.prompt(
            "Which python version to do you want to use?", default=python_version
        )
    )

    packager = None

    gpu = "gpu" in manifest.requirements

    # loading environment
    # check if running in conda
    conda_env = os.environ.get("CONDA_DEFAULT_ENV")
    if conda_env:
        click.echo(f"Found running in conda environment {conda_env}")

        packager = Packager.CONDA

    # Check if pyproject.toml exists
    if os.path.exists("pyproject.toml"):
        # Check if poetry is installed
        try:
            import toml
        except ImportError as e:
            raise ImportError("Please install toml to use this feature") from e

        # Check if poetry is used
        pyproject = toml.load("pyproject.toml")
        if "tool" in pyproject and "poetry" in pyproject["tool"]:
            click.echo("Found poetry project")
            do = True if auto else click.confirm("Do you want to use poetry?")
            if do:
                packager = Packager.POETRY

    if not packager:
        click.echo(
            "No advanced packager found. Considering using poetry with a pyproject.toml or conda with a environment.yml"
        )

        if os.path.exists("requirements.txt"):
            click.prompt("Found requirements.txt file in current directory. Using it.")

        packager = Packager.PIP

    if click.confirm(
        f"Would you like to generate Template Dockerfile: Using {packager} with Python {python_version} {'and' if gpu else 'without'} GPU support"
    ):
        dockfile = build_dockerfile(packager, gpu=gpu, python_version=python_version)
        with open("Dockerfile", "w") as file:
            file.write(dockfile)


def get_builds() -> Dict[str, Build]:
    path = create_arkitekt_folder()
    config_file = os.path.join(path, "builds.yaml")

    builds = {}

    if os.path.exists(config_file):
        with open(config_file, "r") as file:
            config = ConfigFile(**yaml.safe_load(file))

            builds = {build.build_id: build for build in config.builds}
            builds["latest"] = config.latest_build
            return builds
    else:
        return {}


def generate_build(
    builder: str,
    build_id: str,
    manifest: Manifest,
) -> Build:
    path = create_arkitekt_folder()

    config_file = os.path.join(path, "builds.yaml")

    build = Build(
        manifest=manifest,
        builder=builder,
        build_id=build_id,
    )

    if os.path.exists(config_file):
        with open(config_file, "r") as file:
            config = ConfigFile(**yaml.safe_load(file))
            config.builds.append(build)
            config.latest_build = build
    else:
        config = ConfigFile(builds=[build], latest_build=build)

    with open(config_file, "w") as file:
        yaml.safe_dump(
            json.loads(config.json(exclude_none=True, exclude_unset=True)),
            file,
            sort_keys=True,
        )



from arkitekt.apps import Arkitekt
from arkitekt.utils import register, register_structure
from rekuest.actors.reactive.api import log, alog, progress, aprogress
from .builders import easy, qt, publicqt, jupy, scheduler

__all__ = [
    "Arkitekt",
    "register",
    "easy",
    "qt",
    "publicqt",
    "jupy",
    "log",
    "alog",
    "progress",
    "aprogress",
    "scheduler",
    "register_structure",
]



from arkitekt.apps.herre import HerreApp
from pydantic import Field
from arkitekt.healthz import FaktsHealthz
from rath.contrib.fakts.links.aiohttp import FaktsAIOHttpLink
from rath.links.split import SplitLink
from rath.contrib.fakts.links.websocket import FaktsWebsocketLink
from rath.contrib.herre.links.auth import HerreAuthLink
from rekuest.rath import RekuestLinkComposition, RekuestRath
from rekuest.rekuest import Rekuest
from graphql import OperationType
from rekuest.contrib.fakts.websocket_agent_transport import FaktsWebsocketAgentTransport
from rekuest.agents.stateful import StatefulAgent


class ArkitektAgent(StatefulAgent):
    transport: FaktsWebsocketAgentTransport = Field(
        default_factory=lambda: FaktsWebsocketAgentTransport(fakts_group="rekuest.agent")
    )

class ArkitektRekuest(Rekuest):
    rath: RekuestRath = Field(
        default_factory=lambda: RekuestRath(
            link=RekuestLinkComposition(
                auth=HerreAuthLink(),
                split=SplitLink(
                    left=FaktsAIOHttpLink(fakts_group="rekuest"),
                    right=FaktsWebsocketLink(fakts_group="rekuest"),
                    split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
                ),
            )
        )
    )
    agent: StatefulAgent = Field(
        default_factory=lambda: ArkitektAgent()
    )
    healthz: FaktsHealthz = Field(
        default_factory=lambda: FaktsHealthz(fakts_group="rekuest")
    )


class RekuestApp(HerreApp):
    """Mikro App

    It is responsible for setting up the connection to the mikro-server and
    handling authentification and setting up the configuration. Mikro handles the creation of the datalayer and
    the graphql client.

    You can compose this app together with other apps to create a full fledged app. (Like combining with
    arkitekt to enable to call functions that you define on the app). See the example in the docstring.

    Attributes:
        fakts (Fakts): The fakts instance to use.
        mikro (Mikro): The mikro instance to use.
        herre (Herre): The herre instance to use.

    """

    rekuest: ArkitektRekuest = Field(default_factory=ArkitektRekuest)
    """The mikro layer that is used for the datalayer and
    api client
    """



from enum import Enum
import os
import rich_click as click
from .types import Requirement


class Framework(str, Enum):
    VANILLA = "vanilla"
    TENSORFLOW = "tensorflow"
    PYTORCH = "pytorch"


def check_dl_frameworks():
    import site
    import sys

    site_packages = site.getsitepackages()
    site_packages.append(sys.prefix)

    frameworks = {
        "tensorflow": Framework.TENSORFLOW,
        "torch": Framework.PYTORCH,
        "torchvision": Framework.PYTORCH,
        "tensorflow-gpu": Framework.TENSORFLOW,
        "torch-gpu": Framework.PYTORCH,
        "torchvision-gpu": Framework.PYTORCH,
    }

    included_frameworks = set()

    for key, value in frameworks.items():
        for site_package in site_packages:
            if os.path.exists(os.path.join(site_package, key)):
                included_frameworks.add(value)
    return included_frameworks


def inspect_requirements(automatic=False):
    requirements = []
    reasons = []
    frameworks = check_dl_frameworks()
    if frameworks:
        yes = (
            True
            if automatic
            else click.ask(
                "It seems like you are using a deep learning framework. Would you like to require GPU support?",
            )
        )
        if yes:
            requirements.append(Requirement.GPU)
            reasons.append(
                "Deep learning framework detected ('{}')".format(", ".join(frameworks))
            )

    return requirements, reasons



from importlib import import_module
from rich.console import Console
import asyncio
from .ui import construct_run_panel
from .utils import import_builder

console = Console()


async def run_app(entrypoint, app):
    try:
        with console.status("Loading entrypoint module..."):
            try:
                import_module(entrypoint)
            except ModuleNotFoundError as e:
                console.print(f"Could not find entrypoint module {entrypoint}")
                raise e

        console.print("App is running...")
    except asyncio.CancelledError as e:
        console.print("Shutting down...")
        raise e


def import_entrypoint(entrypoint):
    with console.status("Loading entrypoint module..."):
        try:
            import_module(entrypoint)
        except ModuleNotFoundError as e:
            console.print(f"Could not find entrypoint module {entrypoint}")
            raise e


async def run_costum(
    entrypoint, identifier, version, builder: str = "arkitekt.builders.easy"
):
    builder = import_builder(builder)

    import_entrypoint(entrypoint)

    app = builder(identifier, version)

    panel = construct_run_panel(app)
    console.print(panel)

    async with app:
        await app.rekuest.run()


async def run_easy(entrypoint, identifier, version, url, instance_id, headless=False):
    from arkitekt.builders import easy

    import_entrypoint(entrypoint)

    app = easy(identifier, version, url=url, instance_id=instance_id, headless=headless)

    panel = construct_run_panel(app)
    console.print(panel)

    async with app:
        await app.rekuest.run()


async def run_port(entrypoint, identifier, version, url, token, instance_id):
    from arkitekt.builders import port

    import_entrypoint(entrypoint)

    app = port(identifier, version, url=url, token=token, instance_id=instance_id)

    panel = construct_run_panel(
        app,
    )
    console.print(panel)

    async with app:
        await app.rekuest.run()
